==> linstor-server-1.15.0.tar.gz: linstor-server-1.15.0/controller/src/main/java/com/linbit/linstor/core/apicallhandler/controller/autoplacer/StrategyHandler.java <==
package com.linbit.linstor.core.apicallhandler.controller.autoplacer;

import com.linbit.linstor.annotation.SystemContext;
import com.linbit.linstor.api.ApiConsts;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.AutoplaceStrategy.MinMax;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.AutoplaceStrategy.RatingAdditionalInfo;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.Autoplacer.StorPoolWithScore;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.strategies.MaximumFreeSpaceStrategy;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.strategies.MinimumReservedSpaceStrategy;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.strategies.MinimumResourceCountStrategy;
import com.linbit.linstor.core.objects.StorPool;
import com.linbit.linstor.core.repository.SystemConfRepository;
import com.linbit.linstor.logging.ErrorReporter;
import com.linbit.linstor.propscon.Props;
import com.linbit.linstor.security.AccessContext;
import com.linbit.linstor.security.AccessDeniedException;

import javax.inject.Inject;
import javax.inject.Singleton;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

@Singleton
class StrategyHandler
{
    private final List<AutoplaceStrategy> strategies;
    private final SystemConfRepository sysCfgRep;
    private final AccessContext apiCtx;
    private final ErrorReporter errorReporter;

    private final Map<AutoplaceStrategy, Double> dfltWeights;

    @Inject
    StrategyHandler(
        SystemConfRepository sysCfgRepRef,
        @SystemContext AccessContext apiCtxRef,
        MaximumFreeSpaceStrategy freeSpaceStratRef,
        MinimumReservedSpaceStrategy minReservedSpaceStratRef,
        MinimumResourceCountStrategy minRscCountStratRef,
        ErrorReporter errorReporterRef
    )
    {
        strategies = Arrays.asList(freeSpaceStratRef, minReservedSpaceStratRef, minRscCountStratRef);
        sysCfgRep = sysCfgRepRef;
        apiCtx = apiCtxRef;
        errorReporter = errorReporterRef;

        dfltWeights = new HashMap<>();
        for (AutoplaceStrategy strat : strategies)
        {
            dfltWeights.put(strat, 0.0);
        }
        dfltWeights.put(freeSpaceStratRef, 1.0);
    }

    public Collection<StorPoolWithScore> rate(
        Collection<StorPool> storPoolListRef
    )
        throws AccessDeniedException
    {
        RatingAdditionalInfo additionalInfo = new RatingAdditionalInfo();

        Map<AutoplaceStrategy, Double> strategyWeights = getWeights();

        Map<StorPool, StorPoolWithScore> lut = new HashMap<>();
        for (AutoplaceStrategy strat : strategies)
        {
            String stratName = strat.getName();
            double weight = strategyWeights.get(strat);

            Map<StorPool, Double> stratRate = strat.rate(storPoolListRef, additionalInfo);

            double highestValue = Double.NEGATIVE_INFINITY;
            for (Double stratValue : stratRate.values())
            {
                if (highestValue < stratValue && stratValue != 0.0)
                {
                    highestValue = stratValue;
                }
            }

            if (!stratRate.isEmpty())
            {
                errorReporter.logTrace(
                    "Autoplacer.Strategy: Scores of strategy '%s', weight: %f: (raw score, normalized score, weighted final score)",
                    stratName,
                    weight
                );
            }
            for (Entry<StorPool, Double> rate : stratRate.entrySet())
            {
                StorPool sp = rate.getKey();
                StorPoolWithScore prevRating = lut.get(sp);
                double stratValue = rate.getValue();

                if (prevRating == null)
                {
                    prevRating = new StorPoolWithScore(sp, 0);
                    lut.put(sp, prevRating);
                }
                // normalize and weight the value
                double normalizedVal;
                if (highestValue != Double.NEGATIVE_INFINITY)
                {
                    normalizedVal = stratValue / highestValue;
                }
                else
                {
                    normalizedVal = stratValue;
                }
                double normalizdWeightedVal = normalizedVal * weight;
                prevRating.score += normalizdWeightedVal;
                errorReporter.logTrace(
                    "Autoplacer.Strategy: Updated score of StorPool '%s' on Node '%s' to %f (%f, %f, %f)",
                    sp.getName().displayValue,
                    sp.getNode().getName().displayValue,
                    prevRating.score,
                    stratValue,
                    normalizedVal,
                    normalizdWeightedVal
                );
            }
        }

        return lut.values();
    }

    private Map<AutoplaceStrategy, Double> getWeights() throws AccessDeniedException
    {
        Map<AutoplaceStrategy, Double> weights = new HashMap<>(dfltWeights);

        Props ctrlProps = sysCfgRep.getCtrlConfForView(apiCtx);
        Optional<Props> weightsNsOpt = ctrlProps.getNamespace(ApiConsts.NAMESPC_AUTOPLACER_WEIGHTS);
        if (weightsNsOpt.isPresent())
        {
            Props weightsNs = weightsNsOpt.get();
            for (AutoplaceStrategy strat : strategies)
            {
                String stratName = strat.getName();
                String valStr = weightsNs.getProp(stratName);
                double valDouble;
                if (valStr != null)
                {
                    try
                    {
                        valDouble = Double.parseDouble(valStr);
                        errorReporter.logTrace(
                            "Autoplacer.Strategy: Strategy '%s' with weight: %f",
                            stratName,
                            valDouble
                        );
                    }
                    catch (NumberFormatException nfExc)
                    {
                        errorReporter.reportError(
                            nfExc,
                            null,
                            null,
                            "Could not parse '" + valStr + "' for strategy '" + stratName +
                                "'. Defaulting to 1.0"
                        );
                        valDouble = 1.0;
                    }
                }
                else
                {
                    errorReporter.logTrace(
                        "Autoplacer.Strategy: No weight configured for strategy '%s'. Defaulting to 1.0",
                        stratName
                    );
                    valDouble = 1.0;
                }

                if (strat.getMinMax() == MinMax.MINIMIZE)
                {
                    valDouble *= -1;
                }

                weights.put(strat, valDouble);
            }
        }
        return weights;
    }
}

==> linstor-server-1.18.1-rc.1.tar.gz: linstor-server-1.18.1-rc.1/controller/src/main/java/com/linbit/linstor/core/apicallhandler/controller/autoplacer/StrategyHandler.java <==
package com.linbit.linstor.core.apicallhandler.controller.autoplacer;

import com.linbit.linstor.annotation.SystemContext;
import com.linbit.linstor.api.ApiConsts;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.AutoplaceStrategy.MinMax;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.AutoplaceStrategy.RatingAdditionalInfo;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.Autoplacer.StorPoolWithScore;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.strategies.MaximumFreeSpaceStrategy;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.strategies.MinimumReservedSpaceStrategy;
import com.linbit.linstor.core.apicallhandler.controller.autoplacer.strategies.MinimumResourceCountStrategy;
import com.linbit.linstor.core.objects.StorPool;
import com.linbit.linstor.core.repository.SystemConfRepository;
import com.linbit.linstor.logging.ErrorReporter;
import com.linbit.linstor.propscon.Props;
import com.linbit.linstor.security.AccessContext;
import com.linbit.linstor.security.AccessDeniedException;

import javax.inject.Inject;
import javax.inject.Singleton;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

@Singleton
class StrategyHandler
{
    private final List<AutoplaceStrategy> strategies;
    private final SystemConfRepository sysCfgRep;
    private final AccessContext apiCtx;
    private final ErrorReporter errorReporter;

    private final Map<AutoplaceStrategy, Double> dfltWeights;

    @Inject
    StrategyHandler(
        SystemConfRepository sysCfgRepRef,
        @SystemContext AccessContext apiCtxRef,
        MaximumFreeSpaceStrategy freeSpaceStratRef,
        MinimumReservedSpaceStrategy minReservedSpaceStratRef,
        MinimumResourceCountStrategy minRscCountStratRef,
        ErrorReporter errorReporterRef
    )
    {
        strategies = Arrays.asList(freeSpaceStratRef, minReservedSpaceStratRef, minRscCountStratRef);
        sysCfgRep = sysCfgRepRef;
        apiCtx = apiCtxRef;
        errorReporter = errorReporterRef;

        dfltWeights = new HashMap<>();
        for (AutoplaceStrategy strat : strategies)
        {
            dfltWeights.put(strat, 0.0);
        }
        dfltWeights.put(freeSpaceStratRef, 1.0);
    }

    public Collection<StorPoolWithScore> rate(
        Collection<StorPool> storPoolListRef
    )
        throws AccessDeniedException
    {
        RatingAdditionalInfo additionalInfo = new RatingAdditionalInfo();

        Map<AutoplaceStrategy, Double> strategyWeights = getWeights();

        Map<StorPool, StorPoolWithScore> lut = new HashMap<>();
        for (AutoplaceStrategy strat : strategies)
        {
            String stratName = strat.getName();
            double weight = strategyWeights.get(strat);

            Map<StorPool, Double> stratRate = strat.rate(storPoolListRef, additionalInfo);

            double highestValue = Double.NEGATIVE_INFINITY;
            for (Double stratValue : stratRate.values())
            {
                if (highestValue < stratValue && stratValue != 0.0)
                {
                    highestValue = stratValue;
                }
            }

            if (!stratRate.isEmpty())
            {
                errorReporter.logTrace(
                    "Autoplacer.Strategy: Scores of strategy '%s', weight: %f: " +
                        "(raw score, normalized score, weighted final score)",
                    stratName,
                    weight
                );
            }
            for (Entry<StorPool, Double> rate : stratRate.entrySet())
            {
                StorPool sp = rate.getKey();
                StorPoolWithScore prevRating = lut.get(sp);
                double stratValue = rate.getValue();

                if (prevRating == null)
                {
                    prevRating = new StorPoolWithScore(sp, 0);
                    lut.put(sp, prevRating);
                }
                // normalize and weight the value
                double normalizedVal;
                if (highestValue != Double.NEGATIVE_INFINITY)
                {
                    normalizedVal = stratValue / highestValue;
                }
                else
                {
                    normalizedVal = stratValue;
                }
                double normalizdWeightedVal = normalizedVal * weight;
                prevRating.score += normalizdWeightedVal;
                errorReporter.logTrace(
                    "Autoplacer.Strategy: Updated score of StorPool '%s' on Node '%s' to %f (%f, %f, %f)",
                    sp.getName().displayValue,
                    sp.getNode().getName().displayValue,
                    prevRating.score,
                    stratValue,
                    normalizedVal,
                    normalizdWeightedVal
                );
            }
        }

        return lut.values();
    }

    private Map<AutoplaceStrategy, Double> getWeights() throws AccessDeniedException
    {
        Map<AutoplaceStrategy, Double> weights = new HashMap<>(dfltWeights);

        Props ctrlProps = sysCfgRep.getCtrlConfForView(apiCtx);
        Optional<Props> weightsNsOpt = ctrlProps.getNamespace(ApiConsts.NAMESPC_AUTOPLACER_WEIGHTS);
        if (weightsNsOpt.isPresent())
        {
            Props weightsNs = weightsNsOpt.get();
            for (AutoplaceStrategy strat : strategies)
            {
                String stratName = strat.getName();
                String valStr = weightsNs.getProp(stratName);
                double valDouble;
                if (valStr != null)
                {
                    try
                    {
                        valDouble = Double.parseDouble(valStr);
                        errorReporter.logTrace(
                            "Autoplacer.Strategy: Strategy '%s' with weight: %f",
                            stratName,
                            valDouble
                        );
                    }
                    catch (NumberFormatException nfExc)
                    {
                        errorReporter.reportError(
                            nfExc,
                            null,
                            null,
                            "Could not parse '" + valStr + "' for strategy '" + stratName +
                                "'. Defaulting to 1.0"
                        );
                        valDouble = 1.0;
                    }
                }
                else
                {
                    errorReporter.logTrace(
                        "Autoplacer.Strategy: No weight configured for strategy '%s'. Defaulting to 1.0",
                        stratName
                    );
                    valDouble = 1.0;
                }

                if (strat.getMinMax() == MinMax.MINIMIZE)
                {
                    valDouble *= -1;
                }

                weights.put(strat, valDouble);
            }
        }
        return weights;
    }
}
